# wanha - your pesky URL-guardian for IRC
# Copyright (C) 2006-2007 Tommi Saviranta  <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


package Wanha::Score;

use strict;
use warnings FATAL => 'all';
use diagnostics;

use Wanha::DB;


=item add
Add entry to scoreboard

0: nick id
1: chan id
=cut
sub add($$) {
	my ($nick_id, $chan_id) = @_;
	my $update;

	$update = &Wanha::DB::prepare("INSERT INTO scoreboard (nick, chan) " .
			"VALUES (&da&, &db&)");
	$update = &Wanha::DB::set($update,
			"a", $nick_id,
			"b", $chan_id);
	&Wanha::DB::update($update);
}


=item get
Get scoreboard entry

0: nick id
1: chan id
2: auto-add (define to enable)
r: (fame, shame, cheat, wanha)
=cut
sub get($$$) {
	my ($nick_id, $chan_id, $autoadd) = @_;
	my ($query, $result);

	$query = &Wanha::DB::prepare("SELECT fame, shame, cheat, wanha " .
			"FROM scoreboard " .
			"WHERE nick = &da& AND chan = &db&");
	$query = &Wanha::DB::set($query,
			"a", $nick_id,
			"b", $chan_id);
	$result = &Wanha::DB::query($query);
	if (&Wanha::DB::row_count($result) == 1) {
		my ($fame, $shame, $cheat, $wanha) =
			&Wanha::DB::get_row($result);
		return ($fame, $shame, $cheat, $wanha);
	} elsif (&Wanha::DB::row_count($result) == 0) {
		if (defined($autoadd) && $autoadd == 1) {
			&add($nick_id, $chan_id);
			return (0, 0, 0, 0);
		} else {
			return;
		}
	} else {
		printf STDERR "Wanha::Score::get n=%d c=%d " .
				"returned %d matches",
				$nick_id,
				$chan_id,
				&Wanha::DB::row_count($result);
		return;
	}
}


=item get_top
Get TOP-n for channel

0: chan id
1: top-n
2: mode: 0 = fame, 1 = shame, 2 = cheat
r: ( "nick \t score" ... )
=cut
sub get_top($$$) {
	my ($chan_id, $n, $mode) = @_;
	my ($query, $result);
	my @ret;

	if ($mode == 0) {
		$mode = "fame";
	} elsif ($mode == 1) {
		$mode = "shame";
	} elsif ($mode == 2) {
		$mode = "cheat";
	} else {
		die "Wanha::Score::get_top: bad mode, " .
				"chan=$chan_id n=$n mode=$mode";
	}

	$query = &Wanha::DB::prepare(
			"SELECT nicks.nick, scoreboard.$mode " .
			"FROM nicks, scoreboard, chans " .
			"WHERE chans.id = &da& " .
				"AND scoreboard.chan = chans.id " .
				"AND scoreboard.nick = nicks.id " .
			"ORDER BY scoreboard.fame DESC " .
			"LIMIT &db&");
	$query = &Wanha::DB::set($query,
			"a", $chan_id,
			"b", $n);
	$result = &Wanha::DB::query($query);

	while (my ($nick, $score) = &Wanha::DB::get_row($result)) {
		push(@ret, "$nick\t$score");
	}

	return @ret;
}


=item set_value
Set value

0: nick id
1: chan id
2: value key
3: value
=cut
sub set_value($$$$) {
	my ($nick_id, $chan_id, $key, $value) = @_;
	my $update;

	$update = &Wanha::DB::prepare("UPDATE scoreboard SET $key = &da& " .
			"WHERE nick = &db& AND chan = &dc&");
	$update = &Wanha::DB::set($update,
			"a", $value,
			"b", $nick_id,
			"c", $chan_id);
	&Wanha::DB::update($update);
}


=item set_fame
Set user fame

0: nick id
1: chan id
2: fame
=cut
sub set_fame($$$) {
	my ($nick_id, $chan_id, $fame) = @_;
	&set_value($nick_id, $chan_id, "fame", $fame);
}


=item set_shame
Set user shame

0: nick id
1: chan id
2: shame
=cut
sub set_shame($$$) {
	my ($nick_id, $chan_id, $shame) = @_;
	&set_value($nick_id, $chan_id, "shame", $shame);
}


=item set_cheat
Set user cheat score

0: nick id
1: chan id
2: cheat score
=cut
sub set_cheat($$$) {
	my ($nick_id, $chan_id, $shame) = @_;
	&set_value($nick_id, $chan_id, "cheat", $shame);
}



return 1;
