# wanha - your pesky URL-guardian for IRC
# Copyright (C) 2006-2007 Tommi Saviranta  <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


package Wanha::Session;

use strict;
use warnings FATAL => 'all';
use diagnostics;

use Wanha::Chan;
use Wanha::URL;

use Wanha::Cookie;
use Wanha::HTTP;
use Wanha::DB;

use Wanha::Common;


my $cached_chan_id;
my $cached_chan_name;
my $cached_sid;



=item guess_channel
Guess session "prefered" channel

r: (chan_id, chan_name)
=cut
sub guess_channel() {
	(defined($cached_chan_id)) &&
			return ($cached_chan_id, $cached_chan_name);

	my $chan_raw = &Wanha::HTTP::get("chan");
	$chan_raw = &Wanha::HTTP::post("chan") if (! defined($chan_raw));

	(defined($chan_raw)) || return undef;

	my $chan_id = &Wanha::Chan::get_id_smart($chan_raw);
	(defined($chan_id)) || return undef;

	my $chan_name = $chan_raw;
	if ($chan_id eq $chan_raw) {
		$chan_name = &Wanha::Chan::get_name($chan_id);
	}
	$cached_chan_id = $chan_id;
	$cached_chan_name = $chan_name;

	return ($chan_id, $chan_name);
}


=item save_channel
Save session prefered channel

0: sid
1: chan id
=cut
sub save_channel($$) {
	my ($sid, $chan) = @_;

	(defined($sid)) || return;
	($chan) = &guess_channel() if (! defined($chan));
	(! defined($chan) || $chan == -1) && return;

	my %info = &Wanha::Cookie::get_info($sid);
	(defined($info{'chan'}) && $info{'chan'} == $chan) && return;

	&Wanha::Cookie::set_channel($sid, $chan);
}


=item get_id
Get session ID. Also update cookie if necessary.

r: sid
=cut
sub get_id() {
	(defined($cached_sid)) && return $cached_sid;

	my $sid;
	foreach my $line (&Wanha::HTTP::get_cookies()) {
		my ($key, $val) = $line =~ /(.*?)=(.*)/;
		($key ne "sid") && next;
		$sid = $val;
		last;
	}

	$sid = &Wanha::HTTP::get("sid") if (! defined($sid));
	(defined($sid)) || return undef;

	my %info = &Wanha::Cookie::get_info($sid);
	my $new_ip = &Wanha::HTTP::get_ip();

	if ($info{'age_now'} >= $info{'age'} / 2) {
		&Wanha::HTTP::set_cookie("sid", $info{'age'});
		&Wanha::Cookie::update_set_date($sid);
	}

	my $set_ip = (defined($new_ip) && $info{'ip'} ne $new_ip)
			? $new_ip : undef;
	&Wanha::Cookie::set_visits($sid, $info{'visits'} + 1, $set_ip);

	$cached_sid = $sid;

	return $sid;
}


=item check_sid
Check if session ID is ok, send redirect if not.

0: sid
1: return url
=cut
sub check_sid($) {
	my ($sid, $ret) = @_;
	(defined($sid)) && return;
	(defined(&Wanha::Cookie::get_info($sid))) && return;

	&Wanha::HTTP::redirect($ret);
	&Wanha::HTTP::print_header();
	exit(0);
}


=item create_sid
Create new session ID

r: session id
=cut
sub create_sid() {
	my $ip = &Wanha::HTTP::get_ip();
	my $url_id = &Wanha::Common::max(0,
			&Wanha::URL::get_latest_id() - 100);
	return &Wanha::Cookie::create_new($ip, $url_id);
}


=item print_links
Print session links

0: chan
1: return url
=cut
sub print_links($$) {
	my ($chan_id, $ret) = @_;
	my $sid = &get_id();
	
	print "<div id=\"links\">\n";
	print "\t<ul>\n";

	# Channel
	#
	my $chan_name;
	if (defined($sid) && ! defined($chan_id)) {
		my %settings = &Wanha::Cookie::get_info($sid);
		$chan_id = $settings{'chan'};
	}
	if (! defined($chan_id)) {
		($chan_id, $chan_name) = &guess_channel();
	}

	if (defined($chan_id) && $chan_id != -1) {
		my $chan_name = &Wanha::Chan::get_name($chan_id);
		printf "\t\t<li>%s</li>\n", &Wanha::HTML::htmlize($chan_name);
		print "\t\t<ul>\n";
		printf "\t\t\t<li><a href=\"%s\">%s</a></li>\n",
				&Wanha::Link::chan($chan_id),
				&Wanha::Lang::get("session-summary");
		printf "\t\t\t<li><a href=\"%s\">%s</a></li>\n",
				&Wanha::Link::chanstat($chan_id),
				&Wanha::Lang::get("session-stats");
		print "\t\t</ul>\n";
	}

	# General
	#
	printf "\t\t<li>%s</li>\n", &Wanha::Lang::get("session-general");
	print "\t\t<ul>\n";
	printf "\t\t\t<li><a href=\"%s\">%s</a></li>\n",
			&Wanha::Link::stat($chan_id),
			&Wanha::Lang::get("session-stats");
	print "\t\t</ul>\n";

	# Session
	#
	printf "\t\t<li>%s</li>\n", &Wanha::Lang::get("session-session");
	print "\t\t<ul>\n";
	if (defined($sid)) {
		printf "\t\t\t<li><a href=\"%s\">%s</a></li>\n",
				&Wanha::Link::latest(),
				&Wanha::Lang::get("session-latest");
		printf "\t\t\t<li><a href=\"%s\">%s</a></li>\n",
				&Wanha::Link::settings(),
				&Wanha::Lang::get("session-settings");
	} else {
		printf "\t\t\t<li><a href=\"%s\">%s</a></li>\n",
				&Wanha::Link::session(1, $ret),
				&Wanha::Lang::get("session-create");
	}
	print "\t\t</ul>\n";

	print "\t</ul>\n";
	print "</div>\n";
}


return 1;
