#! /usr/bin/perl
#
# wanha - your pesky URL-guardian for IRC
# Copyright (C) 2006-2007 Tommi Saviranta  <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


use strict;
use warnings FATAL => 'all';
use diagnostics;

use Wanha::Chan;

use Wanha::URLList;
use Wanha::Session;

use Wanha::HTML;
use Wanha::HTTP;
use Wanha::DB;

use Wanha::Link;
use Wanha::Common;
use Wanha::Lang;


=item main
Main function
=cut
sub main() {
	my $id = &Wanha::HTTP::get_numeric("chan", 1);
	my $name = &Wanha::Chan::get_name($id);

	# Update cookie
	my $sid = &Wanha::Session::get_id();

	if (! defined($name)) {
		&Wanha::HTTP::redirect(&Wanha::Link::chan(-1));
		&Wanha::HTTP::print_header();
		exit(0);
	}

	&Wanha::HTTP::print_header();
	&Wanha::HTML::print_header();

	&Wanha::Session::save_channel($sid, $id);
	&Wanha::Session::print_links($id, &Wanha::Link::chanstat($id));

	&Wanha::Lang::printf("<h1><&chanstat-for&></h1>\n", $name);

	&Wanha::Lang::printf("<h2><&chanstat-hourly&></h2>\n");
	&hourly($id);

	&Wanha::Lang::printf("<h2><&chanstat-weekdaily&></h2>\n");
	&weekdaily($id);

	&Wanha::Lang::printf("<h2><&chanstat-misc&></h2>\n");
	&urlstat($id);

	&Wanha::Lang::printf("<h2><&chanstat-domain-days&></h2>\n", 30);
	&domain($id, 30, 10);

	&Wanha::Lang::printf("<h2><&chanstat-alltime&></h2>\n");
	&alltime($id, 10);

	&Wanha::Lang::printf(
			"<p><&link-go-to&><a href=\"%s\">" .
			"\u<&link-chan&></a></p>\n",
			&Wanha::Link::chan($id));

	&Wanha::HTML::print_footer(&Wanha::Common::process_time());
}


=item urlsta
Print short channel statistics

0: chan id
=cut
sub urlstat($) {
	my ($chan_id) = @_;
	my ($query, $result);

	# url count
	$query = &Wanha::DB::prepare(
			"SELECT COUNT(id) FROM urls WHERE chan = &da&");
	$query = &Wanha::DB::set($query, "a", $chan_id);
	$result = &Wanha::DB::query($query);
	my ($uniq_url_count) = &Wanha::DB::get_row($result);

	# wanha count
	$query = &Wanha::DB::prepare(
			"SELECT COUNT(id), SUM(count) FROM urls " .
			"WHERE count > 1 AND chan = &da&");
	$query = &Wanha::DB::set($query, "a", $chan_id);
	$result = &Wanha::DB::query($query);
	my ($url_count, $wanha_count) = &Wanha::DB::get_row($result);

	$wanha_count -= $url_count;
	my $total_url_count = $uniq_url_count + $wanha_count;
	&Wanha::Lang::printf("<p><&chanstat-urls-wanha-perc&></p>\n",
			$total_url_count,
			$wanha_count,
			($wanha_count * 100 / $total_url_count));
}


=item hourly($$$)
Print hourly statistics

0: chan id
1: width = 12
2: height = 64
=cut
sub hourly($$$) {
	my ($chan_id, $width, $height) = @_;
	my ($query, $result);
	$width = 16 if (! defined($width));
	$height = 64 if (! defined($height));

	$query = &Wanha::DB::prepare(
			"SELECT DISTINCT hour, COUNT(id) FROM (" .
				"SELECT TO_CHAR(time, 'hh24') AS hour, " .
				"id FROM urls WHERE chan = &da&" .
			") AS hour " .
			"GROUP BY hour ORDER BY hour");
	$query = &Wanha::DB::set($query, "a", $chan_id);
	$result = &Wanha::DB::query($query);

	my @col;
	my @data;
	my $hi = 1;

	while (my ($hour, $count) = &Wanha::DB::get_row($result)) {
		$hour =~ s/^0//;
		while ($hour > $#data + 1) {
			push(@data, 0);
			push(@col, $#col + 1);
		}
		push(@data, $count);
		push(@col, $#col + 1);
		$hi = $count if ($count > $hi);
	}

	&print_graph($#data + 1, $width, $height, $hi, @data, @col);
}


=item weekdaily
Print weekdaily statistics

0: chan id
1: width = 12
2: height = 64
=cut
sub weekdaily($$$) {
	my ($chan_id, $width, $height) = @_;
	my ($query, $result);
	$width = 16 if (! defined($width));
	$height = 64 if (! defined($height));

	# TODO: start with <day>
	$query = &Wanha::DB::prepare(
			"SELECT DISTINCT day, COUNT(id) FROM (" .
				"SELECT TO_CHAR(time, 'd') AS day, " .
				"id FROM urls WHERE chan = &da&" .
			") AS hour " .
			"GROUP BY day ORDER BY day");
	$query = &Wanha::DB::set($query, "a", $chan_id);
	$result = &Wanha::DB::query($query);

	my @order = (2, 3, 4, 5, 6, 7, 1);
	my @col = ( &Wanha::Lang::get("short-mon"),
			&Wanha::Lang::get("short-tue"),
			&Wanha::Lang::get("short-wed"),
			&Wanha::Lang::get("short-thu"),
			&Wanha::Lang::get("short-fri"),
			&Wanha::Lang::get("short-sat"),
			&Wanha::Lang::get("short-sun") );
	my @data = ( 0, 0, 0, 0, 0, 0, 0 );
	my $hi = 1;

	while (my ($day, $count) = &Wanha::DB::get_row($result)) {
		$data[($day - 1)] = $count;
		$hi = $count if ($count > $hi);
	}

	&print_graph(7, $width, $height, $hi, @data, @col);
}


=item alltime
Print all-time statistics

0: chan id
1: n
2: seek (optional)
=cut
sub alltime($$$) {
	my ($chan_id, $n, $seek) = @_;
	my @urls;

	@urls = &get_alltime($chan_id, $n, $seek);

	&Wanha::URLList::print(\@urls, "url", "nick", "date", "count");
}


=item get_alltime
Get all-time statistics

0: chan
1: n
2: seek
r: array of hash (url_id, url_url, url_date, url_count, nick_id, nick_name)
=cut
sub get_alltime($$$) {
	my ($chan_id, $n, $seek) = @_;
	my ($query, $result);
	my @ret;

	$query = "SELECT urls.id, urls.url, " .
		&Wanha::DB::get_date_fmt("urls.time") . ", " .
			"urls.count, nicks.nick, nicks.id " .
		"FROM urls, nicks " .
		"WHERE urls.count > 1 AND urls.chan = &da& " .
			"AND urls.nick = nicks.id " .
		"ORDER BY urls.count DESC, urls.time ASC " .
		"LIMIT &db&";
	$query .= " OFFSET &dc&" if (defined($seek));
	$query = &Wanha::DB::prepare($query);
	$query = &Wanha::DB::set($query,
			"a", $chan_id,
			"b", $n);
	$query = &Wanha::DB::set($query, "c", $seek) if (defined($seek));

	$result = &Wanha::DB::query($query);
	(&Wanha::DB::row_count($result) == 0) && return ();

	while (my ($url_id, $url_url, $url_date, $url_count,
				$nick_name, $nick_id) =
			&Wanha::DB::get_row($result)) {
		my %t = (
			'url_id' => $url_id,
			'url_url' => $url_url,
			'url_date' => $url_date,
			'url_count' => $url_count,
			'nick_id' => $nick_id,
			'nick_name' => $nick_name
			);
		push(@ret, \%t);
	}
	return @ret;
}


=item domain
Print domain statistics

0: chan id
1: days
2: count
=cut
sub domain($$$) {
	my ($chan_id, $days, $count) = @_;
	my ($query, $result);

	$query = &Wanha::DB::prepare(
			"SELECT url FROM urls WHERE chan = &da& AND " .
			"CURRENT_TIMESTAMP - time < INTERVAL '&db& days'");
	$query = &Wanha::DB::set($query,
			"a", $chan_id,
			"b", $days);
	$result = &Wanha::DB::query($query);
	if (&Wanha::DB::row_count($result) == 0) {
		&Wanha::Lang::printf("<p><&chanstat-no-urls&></p>\n");
		return;
	}

	my $url_count = &Wanha::DB::row_count($result);
	my %domains;
	while (my ($url) = &Wanha::DB::get_row($result)) {
		my ($domain) = "$url/" =~ /.*?:\/\/(.*?)\//;
		$domain =~ s/www\.//;
		$domains{$domain}++;
	}

	my @top_domain;
	while (my ($domain, $count) = each %domains) {
		push(@top_domain, "$count\t$domain");
	}
	@top_domain = sort Wanha::Common::compare_rank_desc @top_domain;

	my $l_domain = &Wanha::Lang::get("col-domain");
	my $l_count = &Wanha::Lang::get("col-times-pasted");
	my $l_of_all = &Wanha::Lang::get("col-share-of-all");
	print <<EOF;
<table class="domains">
	<tr>
		<th>$l_domain</th>
		<th>$l_count</th>
		<th>$l_of_all</th>
	</tr>
EOF
	my $d_sum = 0;
	my $max_n = &Wanha::Common::min($#top_domain + 1, $count);
	for (my $i = 0; $i < $max_n; $i++) {
		my ($p_count, $p_domain) = split(/\t/, $top_domain[$i], 2);
		my $p_perc = sprintf("%.2f", $p_count / $url_count * 100);
		my $class = $i % 2 == 1 ? " class=\"zebra\"" : "";
		$d_sum += $p_count;
		print <<EOF;
	<tr$class>
		<td>$p_domain</td>
		<td>$p_count</td>
		<td>$p_perc%</td>
	</tr>
EOF
	}
	my $l_total = &Wanha::Lang::get("col-total");
	my $d_perc = sprintf("%.2f", $d_sum / $url_count * 100);
	print <<EOF;
	<tr class="sum">
		<td>$l_total</th>
		<td>$d_sum</td>
		<td>$d_perc%</td>
	</tr>
</table>
EOF
}


=item print_graph
Print "graph" for statistics

0: count
1: width
2: height
3: highest
@: data
@: cols
=cut
sub print_graph($$$$@@) {
	my $n = shift;
	my $width = shift;
	my $height = shift;
	my $hi = shift;

	print <<EOF;
<table class="graph">
	<tr valign="bottom" class="ul">
EOF
	for (my $i = 0; $i < $n; $i++) {
		my $count = $_[$i];
		my $th = int($count / $hi * $height);
		my $dim = "width=\"$width\" height=\"$th\"";
		my $class = $i == $n - 1 ? " class=\"nol\"" : "";
		print "\t\t<td$class><img src=\"blue24.png\" alt=\"\" " .
				"$dim /></td>\n";
	}
	print "\t</tr>\n\t<tr>\n";
	for (my $i = 0; $i < $n; $i++) {
		my $class = $i == $n - 1 ? " class=\"nol\"" : "";
		print "\t\t<td$class>" . $_[($i + $n)] . "</td>\n";
	}
	print "\t</tr>\n</table>\n";
}


&main();
exit(0);
