#! /usr/bin/perl
#
# wanha - your pesky URL-guardian for IRC
# Copyright (C) 2006-2007 Tommi Saviranta  <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


use strict;
use warnings FATAL => 'all';
use diagnostics;
use Socket;
use Pg;

use Wanha::DB;
use Wanha::Chan;
use Wanha::Nick;
use Wanha::Score;
use Wanha::URL;

use Wanha::Config;


my $bot_pass = $Wanha::Config::bot_pass;
my $irc_debug = $Wanha::Config::irc_debug;

my $irc_server = 0;	# Server index for server cycling

my $irc_cnick;

my $last_msg = time();
my $socket = -1;

$SIG{"INT"} = "graceful_exit";
$SIG{"HUP"} = "read_config";


if ($#ARGV >= 0) {
	&learn_logs();
	exit 0;
}


&read_config();



while (1) {
	($socket == -1) && &get_connected();
	($socket == -1) && next;

	my $buf;
	my $rmask = "";
	my $gotdata = time();
	my $sent_ping = 0;
	my $server_id;

	while ($socket != -1) {
		my $pingtime = time() - $gotdata;
		if ($pingtime >= 30 && $sent_ping == 0) {
			if (defined($server_id)) {
				&print_irc("PING $server_id");
			} else {
				&print_irc("PRIVMSG");
			}
			$sent_ping = 1;
		}
		if ($pingtime >= 120) {
			print "Server not responding!\n";
			last;
		}

		vec($rmask, fileno($socket), 1) = 1;
		my ($nfound, $nrmask) = select($rmask, "", "", 1);
		($nfound <= 0) && next;

		my $tbuf;
		my $nread = sysread($socket, $tbuf, 1024);
		if ($nread <= 0) {
			$socket = -1;
			last;
		}
		$buf .= $tbuf;
		$gotdata = time();
		$sent_ping = 0;

		while (1) {
			my ($in, $rem) = $buf =~ /(.*?)\n(.*)/s;
			(defined($in)) || last;
			$buf = $rem;

		$_ = $in;
		chop;
		$in = $_;
		($irc_debug & 1) && print "<< $in\n";

		($server_id) = /:(.*?) 001/ if (! defined($server_id));

		my ($src, $cmd, $p0, $p1) = split(/ /, $in, 4);
		(defined($src) && ($irc_debug & 4)) && print "src: \"$src\"\n";
		(defined($cmd) && ($irc_debug & 4)) && print "cmd: \"$cmd\"\n";
		(defined($p0) && ($irc_debug & 4)) && print "p0 : \"$p0\"\n";
		(defined($p1) && ($irc_debug & 4)) && print "p1 : \"$p1\"\n";

		if ($cmd eq "PRIVMSG") {
			$p1 =~ s/^://;
			&parse_privmsg($p1, $src, $p0);
		} elsif ($cmd eq "433") {
			$irc_cnick .= "_";
			&print_irc("NICK $irc_cnick");
		} elsif ($cmd eq "003") {
			print "join! $#Wanha::Config::irc_chans\n";
			&join_channels(@Wanha::Config::irc_chans);
		} elsif ($src eq "PING") {
			&print_irc("PONG $cmd");
		}
		}
	}

	print "Disconnected!\n";
	close($socket);
	$socket = -1;
	sleep(10);
}



=item get_connected
Get connected to server
=cut
sub get_connected() {
	my $server_count = $#Wanha::Config::irc_servers + 1;
	$irc_server = 0 if ($irc_server >= $server_count);
	my ($irc_host, $irc_port, $irc_pass) =
			&Wanha::Config::irc_server($irc_server);
	my $t;
	$irc_server++;
	my @t;

	# TODO Make sure this works with IPv6!

	print "Connecting to $irc_host:$irc_port...\n";

	local $SIG{ALRM} = sub {
		print "Connect timed out!\n"; $socket = -1;
	};
	alarm(10); # TODO configure me

	@t = getprotobyname("tcp") or die "getprotobyname: $!";
	my $proto = $t[2];

	my $port = $irc_port;

	if (! (@t = gethostbyname($irc_host))) {
		print "gethostbyname failed: $!\n";
		sleep(5);
		return;
	}
	my $addr = $t[4];
	my $conn = pack("S n a4 x8", AF_INET, $port, $addr);

	my $s;
	if (! socket($s, AF_INET, SOCK_STREAM, $proto)) {
		print "Cannot create socket: $!\n";
		sleep(5);
		return;
	}

	if (! connect($s, $conn)) {
		print "Can't connect to $irc_host:$irc_port!\n";
		sleep(10);
		$socket = -1;
		return;
	}

	select((select($s), $| = 1)[0]); # no buffering
	$socket = $s;

	alarm(0);

	(defined($irc_pass)) && &print_irc("PASS $irc_pass");
	&print_irc("NICK $Wanha::Config::irc_nick");
	&print_irc("USER $Wanha::Config::irc_user 0 0 " .
			":$Wanha::Config::irc_real");

	$irc_cnick = $Wanha::Config::irc_nick;
}


=item join_channels
Join channels

@: Channels
=cut
sub join_channels() {
	print "chans: $#_ .. @_\n";
	foreach (@_) {
		&print_irc("JOIN $_");
	}
}


=item parse_privmsg
Parse incoming private message

0: message
1: src
2: target
=cut
sub parse_privmsg($$$) {
	my ($msg, $src, $chan) = @_;
	my $url;

	($src) = $src =~ /^:(.*?)!/;

	if ($msg =~ /^$irc_cnick[,:]? /i) {
		my ($stuff) = $msg =~ /^$irc_cnick[,:]? (.*)/;
		if ($stuff =~ /^scores/i) {
			if ($chan eq $irc_cnick) {
				&print_top($src);
			} else {
				&print_top($chan);
			}
			return;
		}
		elsif (($stuff =~ /^stfu$/i) || ($stuff =~ /^quiet$/i)) {
			my $nick_id = &Wanha::Nick::get_id($src);
			&Wanha::Nick::set_mute($nick_id, 1);
			print $socket "PRIVMSG $chan :$src, ok\n";
		}
		elsif (($stuff =~ /^speak$/i) || ($stuff =~ /^verbose$/i)) {
			my $nick_id = &Wanha::Nick::get_id($src);
			&Wanha::Nick::set_mute($nick_id, 0);
			print $socket "PRIVMSG $chan :$src, ok\n";
		}
		elsif ($stuff =~ /^learn $bot_pass: /i) {
			&teach_url($msg, $src, $chan);
			return;
		}
	}

	&parse_msg($msg, $src, $chan, undef, $chan);
}


=item parse_msg
TODO: Parse message

0: message
1: src (who pasted)
2: chan (where pasted)
3: date (when pasted) or undef if "now"
4: echo
=cut
sub parse_msg($$$$$) {
	my ($msg, $src, $target, $date, $echo) = @_;
	my $url;
	my $bu_msg = $msg;

	while ($msg =~ /[a-z]+:\/\/[^\s]+/i) {
		my ($b) = $msg =~ /(.)?[a-z]+:\/\//i;
		my $del = " ";
		if ($b =~ /['"]/) {
			$del = $b;
		} elsif ($b =~ /\(/) {
			$del = ")";
		}
		$msg =~ s/\Q$del\E?([a-z]+:\/\/.*?)\Q$del\E(.*)/ $1 $2/i;
		($url, $msg) = $msg =~ /([a-z]+:\/\/[^\s]+)\s*(.*)/i;
		(defined($url)) || next;

		# remove full stop or comma, it hardly ever belongs there
		$url =~ s/[.,]$//;

		my $claim_w = (($bu_msg =~ /^w[:!] \Q$url\E/) ||
				($bu_msg =~ /\sw[:!] \Q$url\E/)) ? 1 : 0;
		&check_url($url, $target, $src, $date, $echo, $claim_w);
	}
}


=item teach_url
Teach URL

0: msg
1: user
2: me
=cut
sub teach_url($$$) {
	my ($msg, $user, $me) = @_;
	my ($chan, $date, $who, $line) = $msg =~
			/learn $bot_pass: (.*?) (.*?) <([^> ]+)> (.*)/;
	
	$date = &convert_date($date);
	
	&parse_msg($line, $who, $chan, $date, $user);

	print $socket "PRIVMSG $user :done\n";
}


=item check_url
Check URL and act accordingly

0: url
1: where it was pasted
2: who pasted
3: time url was pasted
4: where to report if old
5: claim to be wanha
=cut
sub check_url($$$$$$) {
	my ($url, $chan_name, $nick_name, $time, $echo, $claim_w) = @_;

	($url =~ /^\Q$Wanha::Config::base_url\E/) && return;

	$chan_name =~ s/[^\#A-Z0-9\-.()\[\]~_^\%]/-/ig;
	$nick_name =~ s/[^\#A-Z0-9\-.()\[\]~_^\%]/-/ig;
	$url =~ s/(['"\x80-\xff])/sprintf("%%%02X", ord($1))/seg;

	my $chan_id = &Wanha::Chan::get_id($chan_name, 1);
	my ($url_id, $url_count, $url_time, $nick_id, $url_ignore) =
			&Wanha::URL::get_info_by_url($url, $chan_id);

	# ignore certain urls
	(defined($url_id) && $url_ignore != 0) && return;
	
	if (defined($url_id) && $claim_w == 0) {
		# wanha with score
		my $wanha_id = &Wanha::Nick::get_id($nick_name, 1);
		($wanha_id == $nick_id) && return;

		&wanha($url_id, $chan_id, $nick_id,
				$wanha_id, $nick_name, $echo);
	} elsif (! defined($url_id) && $claim_w == 1) {
		# cheating
		&cheat($nick_name, $chan_id, $echo);
	} elsif (defined($url_id) && $claim_w == 1) {
		# pasting known wanha
		# nick_id == original paster id
		# nick_name == wanha paster
		&known_wanha($url_id, $chan_id, $nick_id, $nick_name);
	}
	if (! defined($url_id)) {
		# pasted new url
		&add_url($url, $chan_id, $nick_name, $time);
	}
}


=item wanha
Handle wanha URL

0: url id
1: chan id
2: original paste nick id
3: wanha paste nick id
4: wanha paste nick name
5: echo
=cut
sub wanha($$$$$$) {
	my ($url_id, $chan_id, $orig_id, $wanha_id, $wanha_name, $echo) = @_;
	my ($query, $result);
	my $t;

	&Wanha::DB::transaction_start();

	# scores
	my ($orig_score, $wanha_score);
	($orig_score, $t) = &Wanha::Scores::get($orig_id, $chan_id, 1);
	(defined($orig_score)) || die "wanha score";
	($t, $wanha_score) = &Wanha::Scores::get($wanha_id, $chan_id, 1);
	(defined($wanha_score)) || die "wanha wscore";

	# orig nick
	my $orig_name = &Wanha::Nick::get_name($orig_id);

	# url info
	my ($url_count, $url_time) = &Wanha::URL::get_info_by_id($url_id);

	# update scores
	$orig_score++;
	$wanha_score++;
	&Wanha::Score::set_fame($orig_id, $chan_id, $orig_score);
	&Wanha::Score::set_shame($wanha_id, $chan_id, $wanha_score);

	# update url
	$url_count++;
	&Wanha::URL::set_count($url_id, $url_count);

	# record wanha
	&Wanha::Wanha::add($orig_id, $wanha_id, $chan_id, $url_id);

	&Wanha::DB::transaction_end();

	# announce
	my $wanha_mute = &Wanha::Nick::get_mute($wanha_id);
	my $orig_mute = &Wanha::Nick::get_mute($orig_id);

	if (defined($echo) && $wanha_mute == 0 && $orig_mute == 0) {
		&print_irc("PRIVMSG $echo :$wanha_name, wanha! " .
			"$orig_name scores! " .
			"First pasted: $url_time", 1);
	} elsif (defined($echo) && $wanha_mute == 0 && $orig_mute == 1) {
		&print_irc("PRIVMSG $echo :$wanha_name, wanha! " .
			"First pasted: $url_time", 1);
	}
}


=item known_wanha
Someone pasted wanha knowing it was wanha

0: url id
1: chan id
2: original paster id
3: wanha paster nick
=cut
sub known_wanha($$$$) {
	my ($url_id, $chan_id, $orig_id, $wanha_nick) = @_;
	my ($t0, $t1, $t2);
	my $query;

	&Wanha::DB::transaction_start();

	# url info
	my ($url_count) = &Wanha::URL::get_info_by_id($url_id);
	(defined($url_count)) || die "known_wanha url_count";

	# wanha nick id
	my $wanha_id = &Wanha::Nick::get_id($wanha_nick, 1);

	# scores
	my ($wanha_wanha, $orig_fame);
	($t0, $t1, $t2, $wanha_wanha) =
			&Wanha::Scores::get($wanha_id, $chan_id, 1);
	(defined($wanha_wanha)) || die "known_wanha wanha_id";
	($orig_fame) = &Wanha::Scores::get($orig_id, $chan_id, 1);
	(defined($orig_fame)) || die "known_wanha orig_id";

	# update url
	$url_count++;
	&Wanha::URL::set_count($url_id, $url_count);

	# update scoreboard
	$orig_fame++;
	$wanha_wanha++;
	&Wanha::Score::set_fame($orig_id, $chan_id, $orig_fame);
	&Wanha::Score::set_shame($wanha_id, $chan_id, $wanha_wanha);

	&Wanha::DB::transaction_end();
}


=item cheat
Someone pasted new URL but claimed it was wanha

0: nick
1: chan id
2: chan name
=cut
sub cheat($$$) {
	my ($nick_name, $chan_id, $chan_name) = @_;
	my ($query, $result);

	&Wanha::DB::transaction_start();

	# cheater id
	my $nick_id = &Wanha::Nick::get_id($nick_name, 1);
	(defined($nick_id)) || die "cheat nick";

	# scoreboard
	my ($fame, $shame, $cheat) = &Wanha::Scores::get($nick_id, $chan_id, 1);
	(defined($cheat)) || die "cheat score";

	# update score
	$cheat++;
	&Wanha::Score::set_cheat($nick_id, $chan_id, $cheat);

	&Wanha::DB::transaction_end();

	# announce
	(defined($chan_name)) || return;
	my $mute = &Wanha::Nick::get_mute($nick_id);
	if ($mute == 0) {
		&print_irc("PRIVMSG $chan_name :$nick_name, cheater!", 1);
	}
}


=item add_url
Add new URL to databse

0: url
1: chan id
2: nick name
3: timestamp
=cut
sub add_url($$$$) {
	my ($url, $chan_id, $nick_name, $time) = @_;

	my $nick_id = &Wanha::Nick::get_id($nick_name, 1);
	(defined($nick_id)) || die "add_url nick";

	&Wanha::URL::add($url, $chan_id, $nick_id, $time);
}


=item print_top
Print top-5 on channel

0: chan name
=cut
sub print_top($) {
	my ($chan_name) = @_;
	my ($query, $result);

	my $chan_id = &Wanha::Chan::get_id($chan_name);

	my $top = 5;
	my $res;

	my @top;

	@top = &Wanha::Score::get_top($chan_id, $top, 0);
	if ($#top >= 0) {
		$res = "fame top $top:";
		foreach my $line (@top) {
			my ($nick, $score) = split(/\t/, $line);
			$res .= " $nick: $score,";
		}
		$res =~ s/,$//;
	} else {
		$res .= "no fame";
	}

	$res .= " - ";

	@top = &Wanha::Score::get_top($chan_id, $top, 1);
	if ($#top >= 0) {
		$res .= "shame top $top:";
		foreach my $line (@top) {
			my ($nick, $score) = split(/\t/, $line);
			$res .= " $nick: $score,";
		}
		$res =~ s/,$//;
	} else {
		$res .= "no shame";
	}

	&print_irc("PRIVMSG $chan_name :$res", 1);
}


=item print_irc
Send (raw) data to IRC

0: msg
1: safe
=cut
sub print_irc($$) {
	my ($msg, $safe) = @_;
	my $now = time();

	($now - $last_msg < 2 && defined($safe)) && return;
	($irc_debug & 2) && print ">> $msg\n";
	print $socket "$msg\n";
	$last_msg = $now;
}


=item graceful_exit
Exit
=cut
sub graceful_exit() {
	($socket == -1) && exit(0);

	print $socket "QUIT :sigterm caught\n";
	sleep(1);
	exit(0);
}


=item convert_date
Convert logfile date into "standard" date.

0: date
r: date
=cut
sub convert_date($) {
	my ($date) = @_;
	my %month = ( "Jan"=>"01", "Feb"=>"02", "Mar"=>"03", "Apr"=>"04",
			"May"=>"05", "Jun"=>"06", "Jul"=>"07", "Aug"=>"08",
			"Sep"=>"09", "Oct"=>"10", "Nov"=>"11", "Dec"=>"12" );
	
	my ($y, $m, $d, $t) = $date =~ /(\d+) (.*?) (\d+) (.*)/;
	return "$y-" . $month{"$m"} . "-$d $t";
}


=item learn_logs
Parse IRC logs just if it was just happening

@: files
=cut
sub learn_logs(@) {
	print "year? ";
	my $year = <STDIN>;
	print "channel? ";
	my $chan = <STDIN>;
	chop ($year, $chan);

	foreach (@ARGV) {
		open (FILE, "< $_") or die "Cannot read \"$_\"";
		while (<FILE>) {
			chop;
			my ($month, $day, $hour, $minute, $second,
					$who, $message) =
				/(.*?) (.*?) (\d+):(\d+):(\d+) <([^> ]+)> (.*)/;
			(defined($month)) || next;
			my $date = &convert_date("$year $month $day " .
					"$hour:$minute:$second");
			&parse_msg($message, $who, $chan, $date);
		}
		close (FILE);
	}
}


=item read_config
(Re)read configuration, namely channel list.
=cut
sub read_config() {
	print "read config!\n";
	# TODO
}
